#include <iostream>
#include <chrono>
#include <immintrin.h>


/** Ensure ARRAY_SIZE % 8 == 0. */
#define ARRAY_SIZE 800000000


/** Local functions. */
static void add_serial (int32_t *a, int32_t *b, int32_t *res, int n);
static void add_simd (int32_t *a, int32_t *b, int32_t *res, int n);


/**
 *
 * Main entrance.
 * 
 */
int
main (void) {

    /** Generate random integer arrays. */
    int32_t *a = new int32_t[ARRAY_SIZE],
            *b = new int32_t[ARRAY_SIZE];
    for (int i = 0; i < ARRAY_SIZE; ++i) {
        a[i] = rand() % 1000;
        b[i] = rand() % 1000;
    }

    /** Result array. */
    int32_t *res = new int32_t[ARRAY_SIZE];

    /** Invoke serial add & timing. */
    auto ts = std::chrono::high_resolution_clock::now();
    add_serial(a, b, res, ARRAY_SIZE);
    auto te = std::chrono::high_resolution_clock::now();
    std::cout << " Time elapsed: "
              << std::chrono::duration_cast<std::chrono::milliseconds>(te - ts).count()
              << " milliseconds." << std::endl;

    /** Invode SIMD add & timing. Assume ARRAY_SIZE % 8 == 0. */
    ts = std::chrono::high_resolution_clock::now();
    add_simd(a, b, res, ARRAY_SIZE);
    te = std::chrono::high_resolution_clock::now();
    std::cout << " Time elapsed: "
              << std::chrono::duration_cast<std::chrono::milliseconds>(te - ts).count()
              << " milliseconds." << std::endl;

    return 0;
}


/**
 *
 * Serially add two integer arrays.
 * 
 */
static void
add_serial (int32_t *a, int32_t *b, int32_t *res, int n) {
    for (int i = 0; i < n; ++i)
        res[i] = a[i] + b[i];
}


/**
 *
 * Using AVX2 intrinsics to vectorize the add. Assumes N % 8 == 0.
 * 
 */
static void
add_simd (int32_t *a, int32_t *b, int32_t *res, int n) {
    for (int i = 0; i < n; i += 8) {
        __m256i vector_a = _mm256_load_si256((__m256i *) (a + i)),
                vector_b = _mm256_load_si256((__m256i *) (b + i));
        __m256i vector_res = _mm256_add_epi32(vector_a, vector_b);
        _mm256_store_si256((__m256i *) (res + i), vector_res);
    }
}
